from typing import Any, Dict, Optional

from coinbase.constants import API_PREFIX
from coinbase.rest.types.convert_types import (
    CommitConvertTradeResponse,
    CreateConvertQuoteResponse,
    GetConvertTradeResponse,
)


def create_convert_quote(
    self,
    from_account: str,
    to_account: str,
    amount: str,
    user_incentive_id: Optional[str] = None,
    code_val: Optional[str] = None,
    **kwargs,
) -> CreateConvertQuoteResponse:
    """
    **Create Convert Quote**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/convert/quote

    __________

    **Description:**

    Create a convert quote with a specified source currency, target currency, and amount.

    __________

    **Read more on the official documentation:** `Create Convert Quote <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_createconvertquote>`_
    """
    endpoint = f"{API_PREFIX}/convert/quote"

    data = {
        "from_account": from_account,
        "to_account": to_account,
        "amount": amount,
    }

    trade_incentive_metadata = {
        "user_incentive_id": user_incentive_id,
        "code_val": code_val,
    }
    filtered_trade_incentive_metadata = {
        key: value
        for key, value in trade_incentive_metadata.items()
        if value is not None
    }

    if filtered_trade_incentive_metadata:
        data["trade_incentive_metadata"] = filtered_trade_incentive_metadata

    return CreateConvertQuoteResponse(self.post(endpoint, data=data, **kwargs))


def get_convert_trade(
    self, trade_id: str, from_account: str, to_account: str, **kwargs
) -> GetConvertTradeResponse:
    """
    **Get Convert Trade**
    _____________________

    [GET] https://api.coinbase.com/api/v3/brokerage/convert/trade/{trade_id}

    __________

    **Description:**

    Gets a list of information about a convert trade with a specified trade ID, source currency, and target currency.

    __________

    **Read more on the official documentation:** `Get Convert Trade <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getconverttrade>`_
    """
    endpoint = f"{API_PREFIX}/convert/trade/{trade_id}"

    params = {
        "from_account": from_account,
        "to_account": to_account,
    }

    return GetConvertTradeResponse(self.get(endpoint, params=params, **kwargs))


def commit_convert_trade(
    self, trade_id: str, from_account: str, to_account: str, **kwargs
) -> CommitConvertTradeResponse:
    """
    **Commit Convert Trade**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/convert/trade/{trade_id}

    __________

    **Description:**

    Commits a convert trade with a specified trade ID, source currency, and target currency.

    __________

    **Read more on the official documentation:** `Commit Convert Trade <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_commitconverttrade>`_
    """
    endpoint = f"{API_PREFIX}/convert/trade/{trade_id}"

    data = {
        "from_account": from_account,
        "to_account": to_account,
    }

    return CommitConvertTradeResponse(self.post(endpoint, data=data, **kwargs))



from typing import Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import PortfolioType


# Get API Key Permissions
class GetAPIKeyPermissionsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "can_view" in response:
            self.can_view: Optional[bool] = response.pop("can_view")
        if "can_trade" in response:
            self.can_trade: Optional[bool] = response.pop("can_trade")
        if "can_transfer" in response:
            self.can_transfer: Optional[bool] = response.pop("can_transfer")
        if "portfolio_uuid" in response:
            self.portfolio_uuid: Optional[str] = response.pop("portfolio_uuid")
        if "portfolio_type" in response:
            self.portfolio_type: Optional[PortfolioType] = response.pop(
                "portfolio_type"
            )
        super().__init__(**response)



from typing import Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import RatConvertTrade


# Create Convert Quote
class CreateConvertQuoteResponse(BaseResponse):
    def __init__(self, response: dict):
        if "trade" in response:
            self.trade: Optional[RatConvertTrade] = RatConvertTrade(
                **response.pop("trade")
            )
        super().__init__(**response)


# Get Convert Trade
class GetConvertTradeResponse(BaseResponse):
    def __init__(self, response: dict):
        if "trade" in response:
            self.trade: Optional[RatConvertTrade] = RatConvertTrade(
                **response.pop("trade")
            )
        super().__init__(**response)


# Commit Convert Trade
class CommitConvertTradeResponse(BaseResponse):
    def __init__(self, response: dict):
        if "trade" in response:
            self.trade: Optional[RatConvertTrade] = RatConvertTrade(
                **response.pop("trade")
            )
        super().__init__(**response)



from typing import List, Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import Account


# Get Account
class GetAccountResponse(BaseResponse):
    def __init__(self, response: dict):
        if "account" in response:
            self.account: Optional[Account] = Account(**(response.pop("account")))
        super().__init__(**response)


# List Accounts
class ListAccountsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "accounts" in response:
            self.accounts: Optional[List[Account]] = [
                Account(**account) for account in response.pop("accounts")
            ]
        if "has_next" in response:
            self.has_next: Optional[bool] = response.pop("has_next")
        if "cursor" in response:
            self.cursor: Optional[str] = response.pop("cursor")
        if "size" in response:
            self.size: Optional[int] = response.pop("size")
        super().__init__(**response)



import unittest

from requests.exceptions import HTTPError
from requests_mock import Mocker

from coinbase.__version__ import __version__
from coinbase.rest import RESTClient

from ..constants import TEST_API_KEY, TEST_API_SECRET


class RestBaseTest(unittest.TestCase):
    def test_get(self):
        client = RESTClient(api_key=TEST_API_KEY, api_secret=TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts",
                json=expected_response,
            )

            params = {"limit": 2}
            kwargs = {"test_kwarg": "test"}
            accounts = client.get("/api/v3/brokerage/accounts", params, **kwargs)

            captured_request = m.request_history[0]
            captured_query = captured_request.query
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "GET")

            self.assertEqual(captured_query, "limit=2&test_kwarg=test")

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(accounts, expected_response)

    def test_get_public(self):
        client = RESTClient(api_key=TEST_API_KEY, api_secret=TEST_API_SECRET)

        with Mocker() as m:
            expected_response = {"iso": "2022-01-01T00:00:00Z", "epoch": 1640995200}

            with Mocker() as m:
                m.request(
                    "GET",
                    "https://api.coinbase.com/api/v3/brokerage/time",
                    json=expected_response,
                )

                client.get("/api/v3/brokerage/time", public=True)

                captured_request = m.request_history[0]
                captured_headers = captured_request.headers

                self.assertIn("Authorization", captured_headers)
                self.assertTrue("User-Agent" in captured_headers)
                self.assertEqual(
                    captured_headers["User-Agent"],
                    "coinbase-advanced-py/" + __version__,
                )

    def test_get_public_unauthenticated(self):
        client = RESTClient()

        with Mocker() as m:
            expected_response = {"iso": "2022-01-01T00:00:00Z", "epoch": 1640995200}

            with Mocker() as m:
                m.request(
                    "GET",
                    "https://api.coinbase.com/api/v3/brokerage/time",
                    json=expected_response,
                )

                client.get("/api/v3/brokerage/time", public=True)

                captured_request = m.request_history[0]
                captured_headers = captured_request.headers

                self.assertNotIn("Authorization", captured_headers)
                self.assertTrue("User-Agent" in captured_headers)
                self.assertEqual(
                    captured_headers["User-Agent"],
                    "coinbase-advanced-py/" + __version__,
                )

    def test_post(self):
        client = RESTClient(api_key=TEST_API_KEY, api_secret=TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/portfolios",
                json=expected_response,
            )

            data = {"name": "TestName"}
            kwargs = {"test_kwarg": "test"}

            portfolio = client.post("/api/v3/brokerage/portfolios", data=data, **kwargs)

            captured_request = m.request_history[0]
            captured_json = captured_request.json()
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "POST")

            self.assertEqual(captured_json, {"name": "TestName", "test_kwarg": "test"})

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(portfolio, expected_response)

    def test_put(self):
        client = RESTClient(api_key=TEST_API_KEY, api_secret=TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "PUT",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/1234",
                json=expected_response,
            )

            data = {"name": "TestName"}
            kwargs = {"test_kwarg": "test"}

            portfolio = client.put(
                "/api/v3/brokerage/portfolios/1234", data=data, **kwargs
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "PUT")

            self.assertEqual(captured_json, {"name": "TestName", "test_kwarg": "test"})

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(portfolio, expected_response)

    def test_delete(self):
        client = RESTClient(api_key=TEST_API_KEY, api_secret=TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "DELETE",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/1234",
                json=expected_response,
            )

            kwargs = {"test_kwarg": "test"}

            portfolio = client.delete("/api/v3/brokerage/portfolios/1234", **kwargs)

            captured_request = m.request_history[0]
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.method, "DELETE")

            self.assertEqual(captured_request.json(), kwargs)

            self.assertTrue("User-Agent" in captured_headers)
            self.assertEqual(
                captured_headers["User-Agent"], "coinbase-advanced-py/" + __version__
            )
            self.assertTrue("Authorization" in captured_headers)
            self.assertTrue(captured_headers["Authorization"].startswith("Bearer "))

            self.assertEqual(portfolio, expected_response)

    def test_client_error(self):
        client = RESTClient(api_key=TEST_API_KEY, api_secret=TEST_API_SECRET)

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts",
                status_code=400,
            )

            with self.assertRaises(HTTPError):
                client.get("/api/v3/brokerage/accounts")

    def test_server_error(self):
        client = RESTClient(api_key=TEST_API_KEY, api_secret=TEST_API_SECRET)

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts",
                status_code=500,
            )

            with self.assertRaises(HTTPError):
                client.get("/api/v3/brokerage/accounts")



from typing import Any, Dict, List, Optional

from coinbase.constants import API_PREFIX
from coinbase.rest.types.public_types import (
    GetPublicMarketTradesResponse,
    GetPublicProductBookResponse,
    GetPublicProductCandlesResponse,
    GetPublicProductResponse,
    GetServerTimeResponse,
    ListPublicProductsResponse,
)


def get_unix_time(self, **kwargs) -> GetServerTimeResponse:
    """
    **Get Server Time**
    _________________
    [GET] https://api.coinbase.com/api/v3/brokerage/time

    __________

    **Description:**

    Get the current time from the Coinbase Advanced API. This is a public endpoint.

    __________

    **Read more on the official documentation:** `Get Server Time <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getservertime>`_
    """

    endpoint = f"{API_PREFIX}/time"

    return GetServerTimeResponse(self.get(endpoint, public=True, **kwargs))


def get_public_product_book(
    self,
    product_id: str,
    limit: Optional[int] = None,
    aggregation_price_increment: Optional[str] = None,
    **kwargs,
) -> GetPublicProductBookResponse:
    """
    **Get Public Product Book**
    _________________
    [GET] https://api.coinbase.com/api/v3/brokerage/market/product_book

    __________

    **Description:**

    Get a list of bids/asks for a single product. The amount of detail shown can be customized with the limit parameter.

    __________

    **API Key Permissions:**

    This endpoint is public and does not need authentication.

    __________

    **Read more on the official documentation:** `Get Public Product Book <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getpublicproductbook>`_
    """

    endpoint = f"{API_PREFIX}/market/product_book"

    params = {
        "product_id": product_id,
        "limit": limit,
        "aggregation_price_increment": aggregation_price_increment,
    }

    return GetPublicProductBookResponse(
        self.get(endpoint, params=params, public=True, **kwargs)
    )


def get_public_products(
    self,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
    product_type: Optional[str] = None,
    product_ids: Optional[List[str]] = None,
    contract_expiry_type: Optional[str] = None,
    expiring_contract_status: Optional[str] = None,
    get_all_products: bool = False,
    **kwargs,
) -> ListPublicProductsResponse:
    """
    **List Public Products**
    _________________

    [GET] https://api.coinbase.com/api/v3/brokerage/market/products

    __________

    **Description:**

    Get a list of the available currency pairs for trading.

    __________

    **API Key Permissions:**

    This endpoint is public and does not need authentication.

    __________

    **Read more on the official documentation:** `List Public Products
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getpublicproducts>`_
    """
    endpoint = f"{API_PREFIX}/market/products"

    params = {
        "limit": limit,
        "offset": offset,
        "product_type": product_type,
        "product_ids": product_ids,
        "contract_expiry_type": contract_expiry_type,
        "expiring_contract_status": expiring_contract_status,
        "get_all_products": get_all_products,
    }

    return ListPublicProductsResponse(
        self.get(endpoint, params=params, public=True, **kwargs)
    )


def get_public_product(self, product_id: str, **kwargs) -> GetPublicProductResponse:
    """
    **Public Get Product**
    _______________

    [GET] https://api.coinbase.com/api/v3/brokerage/market/products/{product_id}

    __________

    **Description:**

    Get information on a single product by product ID.

    __________

    **API Key Permissions:**

    This endpoint is public and does not need authentication.

    __________

    **Read more on the official documentation:** `Get Public Product
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getpublicproduct>`_
    """
    endpoint = f"{API_PREFIX}/market/products/{product_id}"

    return GetPublicProductResponse(self.get(endpoint, public=True, **kwargs))


def get_public_candles(
    self,
    product_id: str,
    start: str,
    end: str,
    granularity: str,
    limit: Optional[int] = None,
    **kwargs,
) -> GetPublicProductCandlesResponse:
    """
    **Get Public Product Candles**
    __________

    [GET] https://api.coinbase.com/api/v3/brokerage/market/products/{product_id}/candles

    __________

    **Description:**

    Get rates for a single product by product ID, grouped in buckets.

    __________

    **API Key Permissions:**

    This endpoint is public and does not need authentication.

    __________

    **Read more on the official documentation:** `Get Public Product Candles
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getpubliccandles>`_
    """
    endpoint = f"{API_PREFIX}/market/products/{product_id}/candles"

    params = {"start": start, "end": end, "granularity": granularity, "limit": limit}

    return GetPublicProductCandlesResponse(
        self.get(endpoint, params=params, public=True, **kwargs)
    )


def get_public_market_trades(
    self,
    product_id: str,
    limit: int,
    start: Optional[str] = None,
    end: Optional[str] = None,
    **kwargs,
) -> GetPublicMarketTradesResponse:
    """
    **Get Public Market Trades**
    _____________________

    [GET] https://api.coinbase.com/api/v3/brokerage/market/products/{product_id}/ticker

    __________

    **Description:**

    Get snapshot information, by product ID, about the last trades (ticks), best bid/ask, and 24h volume.

    __________

    **API Key Permissions:**

    This endpoint is public and does not need authentication.

    __________

    **Read more on the official documentation:** `Get Public Market Trades
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getpublicmarkettrades>`_
    """
    endpoint = f"{API_PREFIX}/market/products/{product_id}/ticker"

    params = {"limit": limit, "start": start, "end": end}

    return GetPublicMarketTradesResponse(
        self.get(endpoint, params=params, public=True, **kwargs)
    )



from typing import List, Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import PaymentMethod


# List Payment Methods
class ListPaymentMethodsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "payment_methods" in response:
            self.payment_methods: Optional[List[PaymentMethod]] = [
                PaymentMethod(**method) for method in response.pop("payment_methods")
            ]

        super().__init__(**response)


# Get Payment Method
class GetPaymentMethodResponse(BaseResponse):
    def __init__(self, response: dict):
        if "payment_method" in response:
            self.payment_method: Optional[PaymentMethod] = PaymentMethod(
                **response.pop("payment_method")
            )
        super().__init__(**response)



import logging
import os
from multiprocessing import AuthenticationError
from typing import IO, Any, Dict, Optional, Union

import requests
from requests.exceptions import HTTPError

from coinbase import jwt_generator
from coinbase.api_base import APIBase, get_logger
from coinbase.constants import (
    API_ENV_KEY,
    API_SECRET_ENV_KEY,
    BASE_URL,
    RATE_LIMIT_HEADERS,
    USER_AGENT,
)

logger = get_logger("coinbase.RESTClient")


def handle_exception(response):
    """Raises :class:`HTTPError`, if one occurred.

    :meta private:
    """
    http_error_msg = ""
    reason = response.reason

    if 400 <= response.status_code < 500:
        if (
            response.status_code == 403
            and '"error_details":"Missing required scopes"' in response.text
        ):
            http_error_msg = f"{response.status_code} Client Error: Missing Required Scopes. Please verify your API keys include the necessary permissions."
        else:
            http_error_msg = (
                f"{response.status_code} Client Error: {reason} {response.text}"
            )
    elif 500 <= response.status_code < 600:
        http_error_msg = (
            f"{response.status_code} Server Error: {reason} {response.text}"
        )

    if http_error_msg:
        logger.error(f"HTTP Error: {http_error_msg}")
        raise HTTPError(http_error_msg, response=response)


class RESTBase(APIBase):
    """
    :meta private:
    """

    def __init__(
        self,
        api_key: Optional[str] = os.getenv(API_ENV_KEY),
        api_secret: Optional[str] = os.getenv(API_SECRET_ENV_KEY),
        key_file: Optional[Union[IO, str]] = None,
        base_url=BASE_URL,
        timeout: Optional[int] = None,
        verbose: Optional[bool] = False,
        rate_limit_headers: Optional[bool] = False,
    ):
        super().__init__(
            api_key=api_key,
            api_secret=api_secret,
            key_file=key_file,
            base_url=base_url,
            timeout=timeout,
            verbose=verbose,
        )
        self.rate_limit_headers = rate_limit_headers
        self.session = requests.Session()
        if verbose:
            logger.setLevel(logging.DEBUG)

    def get(
        self, url_path, params: Optional[dict] = None, public=False, **kwargs
    ) -> Dict[str, Any]:
        """
        **GET Request**
        _____________________________

        __________

        **Parameters:**

        - **url_path | (str)** - the URL path
        - **params | Optional ([dict])** - the query parameters
        - **public | (bool)** - flag indicating whether to treat endpoint as public


        """

        params = params or {}

        if kwargs:
            params.update(kwargs)

        return self.prepare_and_send_request(
            "GET", url_path, params, data=None, public=public
        )

    def post(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        **Authenticated POST Request**
        ______________________________

        __________

         **Parameters:**

        - **url_path | (str)** - the URL path
        - **params | Optional ([dict])** - the query parameters
        - **data | Optional ([dict])** - the request body
        """
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("POST", url_path, params, data)

    def put(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        **Authenticated PUT Request**
        _____________________________

        __________

        **Parameters:**

        - **url_path | (str)** - the URL path
        - **params | Optional ([dict])** - the query parameters
        - **data | Optional ([dict])** - the request body
        """
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("PUT", url_path, params, data)

    def delete(
        self,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        **Authenticated DELETE Request**
        ________________________________

        __________

        **Parameters:**

        - **url_path | (str)** - the URL path
        - **params | Optional ([dict])** - the query parameters
        - **data | Optional ([dict])** - the request body
        """
        data = data or {}

        if kwargs:
            data.update(kwargs)

        return self.prepare_and_send_request("DELETE", url_path, params, data)

    def prepare_and_send_request(
        self,
        http_method,
        url_path,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
        public=False,
    ):
        """
        :meta private:
        """
        if not self.is_authenticated and not public:
            raise AuthenticationError(
                "Unauthenticated request to private endpoint. If you wish to access private endpoints, you must provide your API key and secret when initializing the RESTClient."
            )

        headers = self.set_headers(http_method, url_path)

        if params is not None:
            params = {key: value for key, value in params.items() if value is not None}

        if data is not None:
            data = {key: value for key, value in data.items() if value is not None}

        return self.send_request(http_method, url_path, params, headers, data=data)

    def send_request(self, http_method, url_path, params, headers, data=None):
        """
        :meta private:
        """
        if data is None:
            data = {}

        url = f"https://{self.base_url}{url_path}"

        logger.debug(f"Sending {http_method} request to {url}")

        response = self.session.request(
            http_method,
            url,
            params=params,
            json=data,
            headers=headers,
            timeout=self.timeout,
        )
        handle_exception(response)  # Raise an HTTPError for bad responses

        logger.debug(f"Raw response: {response.json()}")

        response_data = response.json()

        if self.rate_limit_headers:
            response_headers = dict(response.headers)
            specific_headers = {
                key: response_headers.get(key, None) for key in RATE_LIMIT_HEADERS
            }

            response_data = {**response_data, **specific_headers}

        return response_data

    def set_headers(self, method, path):
        """
        :meta private:
        """
        uri = f"{method} {self.base_url}{path}"

        return {
            "User-Agent": USER_AGENT,
            "Content-Type": "application/json",
            **(
                {
                    "Authorization": f"Bearer {jwt_generator.build_rest_jwt(uri, self.api_key, self.api_secret)}",
                }
                if self.is_authenticated
                else {}
            ),
        }



import base64
import json
import unittest

import jwt

from coinbase import jwt_generator

from .constants import TEST_API_KEY, TEST_API_SECRET


class JwtGeneratorTest(unittest.TestCase):
    def test_build_rest_jwt(self):
        uri = jwt_generator.format_jwt_uri("GET", "/api/v3/brokerage/accounts")
        result_jwt = jwt_generator.build_rest_jwt(uri, TEST_API_KEY, TEST_API_SECRET)

        decoded_data = jwt.decode(result_jwt, TEST_API_SECRET, algorithms=["ES256"])
        header_bytes = base64.urlsafe_b64decode(str(result_jwt.split(".")[0] + "=="))
        decoded_header = json.loads(header_bytes.decode("utf-8"))

        self.assertEqual(decoded_data["sub"], TEST_API_KEY)
        self.assertEqual(decoded_data["iss"], "cdp")
        self.assertEqual(decoded_data["uri"], uri)
        self.assertEqual(decoded_header["kid"], TEST_API_KEY)

    def test_build_ws_jwt(self):
        result_jwt = jwt_generator.build_ws_jwt(TEST_API_KEY, TEST_API_SECRET)

        decoded_data = jwt.decode(result_jwt, TEST_API_SECRET, algorithms=["ES256"])
        header_bytes = base64.urlsafe_b64decode(str(result_jwt.split(".")[0] + "=="))
        decoded_header = json.loads(header_bytes.decode("utf-8"))

        self.assertEqual(decoded_data["sub"], TEST_API_KEY)
        self.assertEqual(decoded_data["iss"], "cdp")
        self.assertNotIn("uri", decoded_data)
        self.assertEqual(decoded_header["kid"], TEST_API_KEY)

    def test_build_rest_jwt_error(self):
        with self.assertRaises(Exception):
            uri = jwt_generator.format_jwt_uri("GET", "/api/v3/brokerage/accounts")
            jwt_generator.build_rest_jwt(uri, TEST_API_KEY, "bad_secret")



import unittest

from requests_mock import Mocker

from coinbase.rest import RESTClient
from coinbase.rest.types.orders_types import ListOrdersResponse

from ..constants import TEST_API_KEY, TEST_API_SECRET


class OrdersTest(unittest.TestCase):
    def test_create_order(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order_configuration = {"market_market_ioc": {"quote_size": "1"}}

            order = client.create_order(
                "client_order_id_1",
                "product_id_1",
                "BUY",
                order_configuration,
                self_trade_prevention_id="self_trade_prevention_id_1",
                margin_type="CROSS",
                leverage="5",
                retail_portfolio_id="portfolio_id_1",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {"market_market_ioc": {"quote_size": "1"}},
                    "self_trade_prevention_id": "self_trade_prevention_id_1",
                    "margin_type": "CROSS",
                    "leverage": "5",
                    "retail_portfolio_id": "portfolio_id_1",
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_market_order(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )

            order = client.market_order(
                "client_order_id_1", "product_id_1", "BUY", quote_size="1"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()
            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {"market_market_ioc": {"quote_size": "1"}},
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_market_order_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )

            order = client.market_order_buy("client_order_id_1", "product_id_1", "1")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {"market_market_ioc": {"quote_size": "1"}},
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_market_order_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )

            order = client.market_order_sell("client_order_id_1", "product_id_1", "1")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {"market_market_ioc": {"base_size": "1"}},
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_ioc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_ioc(
                "client_order_id_1", "product_id_1", "BUY", "1", "100"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "sor_limit_ioc": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_ioc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_ioc_buy(
                "client_order_id_1", "product_id_1", "1", "100"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "sor_limit_ioc": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_ioc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_ioc_sell(
                "client_order_id_1", "product_id_1", "1", "100"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "sor_limit_ioc": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_gtc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_gtc(
                "client_order_id_1",
                "product_id_1",
                "BUY",
                "1",
                "100",
                True,
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "post_only": True,
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_gtc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_gtc_buy(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                True,
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "post_only": True,
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_gtc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_gtc_sell(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                True,
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "limit_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "post_only": True,
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_gtd(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_gtd(
                "client_order_id_1",
                "product_id_1",
                "BUY",
                "1",
                "100",
                "2022-01-01T00:00:00Z",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "end_time": "2022-01-01T00:00:00Z",
                            "post_only": False,
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_gtd_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_gtd_buy(
                "client_order_id_1", "product_id_1", "1", "100", "2022-01-01T00:00:00Z"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "end_time": "2022-01-01T00:00:00Z",
                            "post_only": False,
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_gtd_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_gtd_sell(
                "client_order_id_1", "product_id_1", "1", "100", "2022-01-01T00:00:00Z"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "limit_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "end_time": "2022-01-01T00:00:00Z",
                            "post_only": False,
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_fok(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_fok(
                "client_order_id_1", "product_id_1", "BUY", "1", "100"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_fok": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_fok_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_fok_buy(
                "client_order_id_1", "product_id_1", "1", "100"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_fok": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_limit_order_fok_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.limit_order_fok_sell(
                "client_order_id_1", "product_id_1", "1", "100"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "limit_limit_fok": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_stop_limit_order_gtc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.stop_limit_order_gtc(
                "client_order_id_1",
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_stop_limit_order_gtc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.stop_limit_order_gtc_buy(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                "90",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_stop_limit_order_gtc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.stop_limit_order_gtc_sell(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                "90",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_stop_limit_order_gtd(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.stop_limit_order_gtd(
                "client_order_id_1",
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_stop_limit_order_gtd_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.stop_limit_order_gtd_buy(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_stop_limit_order_gtd_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.stop_limit_order_gtd_sell(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_trigger_bracket_order_gtc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.trigger_bracket_order_gtc(
                "client_order_id_1",
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_trigger_bracket_order_gtc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.trigger_bracket_order_gtc_buy(
                "client_order_id_1", "product_id_1", "1", "100", "90"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_trigger_bracket_order_gtc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.trigger_bracket_order_gtc_sell(
                "client_order_id_1", "product_id_1", "1", "100", "90"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "trigger_bracket_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_trigger_bracket_order_gtd(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.trigger_bracket_order_gtd(
                "client_order_id_1",
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_trigger_bracket_order_gtd_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.trigger_bracket_order_gtd_buy(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_trigger_bracket_order_gtd_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders",
                json=expected_response,
            )
            order = client.trigger_bracket_order_gtd_sell(
                "client_order_id_1",
                "product_id_1",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "trigger_bracket_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                        }
                    },
                },
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_get_order(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/orders/historical/order_id_1",
                json=expected_response,
            )
            order = client.get_order("order_id_1")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(order.__dict__, expected_response)

    def test_list_orders(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"orders": [{"order_id": "1234"}, {"order_id": "5678"}]}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/orders/historical/batch",
                json=expected_response,
            )
            orders = client.list_orders(
                product_ids=["product_id_1", "product_id_2"],
                order_status="OPEN",
                limit=2,
                product_type="SPOT",
            )

            captured_request = m.request_history[0]

            self.assertEqual(
                captured_request.query,
                "product_ids=product_id_1&product_ids=product_id_2&order_status=open&limit=2&product_type=spot",
            )
            actual_response_dict = orders.to_dict()
            expected_response_dict = ListOrdersResponse(expected_response).to_dict()
            self.assertEqual(actual_response_dict, expected_response_dict)

    def test_get_fills(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"orders": [{"order_id": "1234"}]}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/orders/historical/fills",
                json=expected_response,
            )
            orders = client.get_fills(
                order_ids=["1234"],
                product_ids=["product_id_1"],
                retail_portfolio_id="portfolio_id_1",
                limit=2,
                cursor="abc",
            )

            captured_request = m.request_history[0]

            self.assertEqual(
                captured_request.query,
                "order_ids=1234&product_ids=product_id_1&retail_portfolio_id=portfolio_id_1&limit=2&cursor=abc",
            )
            self.assertEqual(orders.__dict__, expected_response)

    def test_edit_order(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "order_id_1"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/edit",
                json=expected_response,
            )
            order = client.edit_order("order_id_1", "100", "50")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json, {"order_id": "order_id_1", "size": "100", "price": "50"}
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_preview_edit_order(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "order_id_1"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/edit_preview",
                json=expected_response,
            )
            order = client.preview_edit_order("order_id_1", "100", "50")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json, {"order_id": "order_id_1", "size": "100", "price": "50"}
            )
            self.assertEqual(order.__dict__, expected_response)

    def test_cancel_orders(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "order_id_1"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/batch_cancel",
                json=expected_response,
            )
            order = client.cancel_orders(["order_id_1", "order_id_2"])

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(captured_json, {"order_ids": ["order_id_1", "order_id_2"]})
            self.assertEqual(order.__dict__, expected_response)

    def test_preview_order(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "order_id_1"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )

            order_configuration = {"market_market_ioc": {"quote_size": "1"}}

            preview = client.preview_order(
                "product_id_1",
                "BUY",
                order_configuration,
                leverage="5",
                margin_type="CROSS",
                retail_portfolio_id="portfolio_id_1",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {"market_market_ioc": {"quote_size": "1"}},
                    "leverage": "5",
                    "margin_type": "CROSS",
                    "retail_portfolio_id": "portfolio_id_1",
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_market_order(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )

            preview = client.preview_market_order("product_id_1", "BUY", quote_size="1")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {"market_market_ioc": {"quote_size": "1"}},
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_market_order_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )

            preview = client.preview_market_order_buy("product_id_1", "1")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {"market_market_ioc": {"quote_size": "1"}},
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_market_order_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )

            preview = client.preview_market_order_sell("product_id_1", "1")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {"market_market_ioc": {"base_size": "1"}},
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_ioc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_ioc("product_id_1", "BUY", "1", "100")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "sor_limit_ioc": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_ioc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_ioc_buy("product_id_1", "1", "100")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "sor_limit_ioc": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_ioc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_ioc_sell("product_id_1", "1", "100")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "sor_limit_ioc": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_gtc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_gtc(
                "product_id_1",
                "BUY",
                "1",
                "100",
                True,
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "post_only": True,
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_gtc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_gtc_buy(
                "product_id_1",
                "1",
                "100",
                True,
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "post_only": True,
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_gtc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_gtc_sell(
                "product_id_1",
                "1",
                "100",
                True,
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "limit_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "post_only": True,
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_gtd(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_gtd(
                "product_id_1",
                "BUY",
                "1",
                "100",
                "2022-01-01T00:00:00Z",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "end_time": "2022-01-01T00:00:00Z",
                            "post_only": False,
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_gtd_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_gtd_buy(
                "product_id_1", "1", "100", "2022-01-01T00:00:00Z"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "end_time": "2022-01-01T00:00:00Z",
                            "post_only": False,
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_gtd_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_gtd_sell(
                "product_id_1", "1", "100", "2022-01-01T00:00:00Z"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "limit_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "end_time": "2022-01-01T00:00:00Z",
                            "post_only": False,
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_fok(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_fok("product_id_1", "BUY", "1", "100")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_fok": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_fok_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_fok_buy("product_id_1", "1", "100")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "limit_limit_fok": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_limit_order_fok_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_limit_order_fok_sell("product_id_1", "1", "100")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "limit_limit_fok": {"base_size": "1", "limit_price": "100"}
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_stop_limit_order_gtc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_stop_limit_order_gtc(
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_stop_limit_order_gtc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_stop_limit_order_gtc_buy(
                "product_id_1",
                "1",
                "100",
                "90",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_stop_limit_order_gtc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_stop_limit_order_gtc_sell(
                "product_id_1",
                "1",
                "100",
                "90",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_stop_limit_order_gtd(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_stop_limit_order_gtd(
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_stop_limit_order_gtd_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_stop_limit_order_gtd_buy(
                "product_id_1",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_stop_limit_order_gtd_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_stop_limit_order_gtd_sell(
                "product_id_1",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
                "STOP_DIRECTION_STOP_UP",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "stop_limit_stop_limit_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                            "stop_direction": "STOP_DIRECTION_STOP_UP",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_trigger_bracket_order_gtc(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_trigger_bracket_order_gtc(
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_trigger_bracket_order_gtc_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_trigger_bracket_order_gtc_buy(
                "product_id_1", "1", "100", "90"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_trigger_bracket_gtc_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_trigger_bracket_order_gtc_sell(
                "product_id_1", "1", "100", "90"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "trigger_bracket_gtc": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_trigger_bracket_order_gtd(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_trigger_bracket_order_gtd(
                "product_id_1",
                "BUY",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_trigger_bracket_order_gtd_buy(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_trigger_bracket_order_gtd_buy(
                "product_id_1",
                "1",
                "100",
                "90",
                "2022-01-01T00:00:00Z",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "BUY",
                    "order_configuration": {
                        "trigger_bracket_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_preview_trigger_bracket_gtd_sell(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"order_id": "1234"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/preview",
                json=expected_response,
            )
            preview = client.preview_trigger_bracket_order_gtd_sell(
                "product_id_1", "1", "100", "90", "2022-01-01T00:00:00Z"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "product_id": "product_id_1",
                    "side": "SELL",
                    "order_configuration": {
                        "trigger_bracket_gtd": {
                            "base_size": "1",
                            "limit_price": "100",
                            "stop_trigger_price": "90",
                            "end_time": "2022-01-01T00:00:00Z",
                        }
                    },
                },
            )
            self.assertEqual(preview.__dict__, expected_response)

    def test_close_position(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {
            "client_order_id": "client_order_id_1",
            "product_id": "product_id_1",
        }

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/orders/close_position",
                json=expected_response,
            )
            closedOrder = client.close_position(
                "client_order_id_1", "product_id_1", "100"
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "client_order_id": "client_order_id_1",
                    "product_id": "product_id_1",
                    "size": "100",
                },
            )
            self.assertEqual(closedOrder.__dict__, expected_response)



from typing import Any, Dict, Optional

from coinbase.constants import API_PREFIX
from coinbase.rest.types.portfolios_types import (
    CreatePortfolioResponse,
    DeletePortfolioResponse,
    EditPortfolioResponse,
    GetPortfolioBreakdownResponse,
    ListPortfoliosResponse,
    MovePortfolioFundsResponse,
)


def get_portfolios(
    self, portfolio_type: Optional[str] = None, **kwargs
) -> ListPortfoliosResponse:
    """
    **List Portfolios**
    ___________________

    [GET] https://api.coinbase.com/api/v3/brokerage/portfolios

    __________

    **Description:**

    Get a list of all portfolios of a user.

    __________

    **Read more on the official documentation:** `List Portfolios
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getportfolios>`_
    """
    endpoint = f"{API_PREFIX}/portfolios"

    params = {"portfolio_type": portfolio_type}

    return ListPortfoliosResponse(self.get(endpoint, params=params, **kwargs))


def create_portfolio(self, name: str, **kwargs) -> CreatePortfolioResponse:
    """
    **Create Portfolio**
    ____________________

    [POST] https://api.coinbase.com/api/v3/brokerage/portfolios

    __________

    **Description:**

    Create a portfolio.

    __________

    **Read more on the official documentation:** `Create Portfolio
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_createportfolio>`_
    """
    endpoint = f"{API_PREFIX}/portfolios"

    data = {
        "name": name,
    }

    return CreatePortfolioResponse(self.post(endpoint, data=data, **kwargs))


def get_portfolio_breakdown(
    self, portfolio_uuid: str, currency: Optional[str] = None, **kwargs
) -> GetPortfolioBreakdownResponse:
    """
    **Get Portfolio Breakdown**
    ___________________________

    [GET] https://api.coinbase.com/api/v3/brokerage/portfolios/{portfolio_uuid}

    __________

    **Description:**

    Get the breakdown of a portfolio by portfolio ID.

    __________

    **Read more on the official documentation:** `Get Portfolio Breakdown
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getportfoliobreakdown>`_
    """
    endpoint = f"{API_PREFIX}/portfolios/{portfolio_uuid}"

    params = {"currency": currency}
    return GetPortfolioBreakdownResponse(self.get(endpoint, params=params, **kwargs))


def move_portfolio_funds(
    self,
    value: str,
    currency: str,
    source_portfolio_uuid: str,
    target_portfolio_uuid: str,
    **kwargs,
) -> MovePortfolioFundsResponse:
    """
    **Move Portfolio Funds**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/portfolios/move_funds

    __________

    **Description:**

    Transfer funds between portfolios.

    __________

    **Read more on the official documentation:** `Move Portfolio Funds
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_moveportfoliofunds>`_
    """
    endpoint = f"{API_PREFIX}/portfolios/move_funds"

    data = {
        "funds": {
            "value": value,
            "currency": currency,
        },
        "source_portfolio_uuid": source_portfolio_uuid,
        "target_portfolio_uuid": target_portfolio_uuid,
    }

    return MovePortfolioFundsResponse(self.post(endpoint, data=data, **kwargs))


def edit_portfolio(
    self, portfolio_uuid: str, name: str, **kwargs
) -> EditPortfolioResponse:
    """
    **Edit Portfolio**
    __________________

    [PUT] https://api.coinbase.com/api/v3/brokerage/portfolios/{portfolio_uuid}

    __________

    **Description:**

    Modify a portfolio by portfolio ID.

    __________

    **Read more on the official documentation:** `Edit Portfolio
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_editportfolio>`_
    """
    endpoint = f"{API_PREFIX}/portfolios/{portfolio_uuid}"

    data = {
        "name": name,
    }

    return EditPortfolioResponse(self.put(endpoint, data=data, **kwargs))


def delete_portfolio(self, portfolio_uuid: str, **kwargs) -> DeletePortfolioResponse:
    """
    **Delete Portfolio**
    ____________________

    [DELETE] https://api.coinbase.com/api/v3/brokerage/portfolios/{portfolio_uuid}

    __________

    **Description:**

    Delete a portfolio by portfolio ID.

    __________

    **Read more on the official documentation:** `Delete Portfolio
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_deleteportfolio>`_
    """
    endpoint = f"{API_PREFIX}/portfolios/{portfolio_uuid}"

    return DeletePortfolioResponse(self.delete(endpoint, **kwargs))



from typing import Any


class BaseResponse:
    def __init__(self, **data):
        for key in list(data.keys()):
            setattr(self, key, data.pop(key))

    def __getitem__(self, key: str) -> Any:
        return self.__dict__.get(key)

    def __repr__(self):
        return str(self.__dict__)

    def to_dict(self) -> dict:
        dict_response = {}
        for key, value in self.__dict__.items():
            if isinstance(value, BaseResponse):
                dict_response[key] = value.to_dict()
            elif isinstance(value, list):
                dict_response[key] = [
                    item.to_dict() if isinstance(item, BaseResponse) else item
                    for item in value
                ]
            else:
                dict_response[key] = value
        return dict_response



from typing import List, Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import Portfolio, PortfolioBreakdown


# List Portfolios
class ListPortfoliosResponse(BaseResponse):
    def __init__(self, response: dict):
        if "portfolios" in response:
            self.portfolios: Optional[List[Portfolio]] = [
                Portfolio(**portfolio) for portfolio in response.pop("portfolios")
            ]
        super().__init__(**response)


# Create Portfolio
class CreatePortfolioResponse(BaseResponse):
    def __init__(self, response: dict):
        if "portfolio" in response:
            self.portfolio: Optional[Portfolio] = Portfolio(**response.pop("portfolio"))
        super().__init__(**response)


# Move Portfolio Funds
class MovePortfolioFundsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "source_portfolio_uuid" in response:
            self.source_portfolio_uuid: Optional[str] = response.pop(
                "source_portfolio_uuid"
            )
        if "target_portfolio_uuid" in response:
            self.target_portfolio_uuid: Optional[str] = response.pop(
                "target_portfolio_uuid"
            )
        super().__init__(**response)


# Get Portfolio Breakdown
class GetPortfolioBreakdownResponse(BaseResponse):
    def __init__(self, response: dict):
        if "breakdown" in response:
            self.breakdown: Optional[PortfolioBreakdown] = PortfolioBreakdown(
                **response.pop("breakdown")
            )
        super().__init__(**response)


# Delete Portfolio
class DeletePortfolioResponse(BaseResponse):
    def __init__(self, response: dict):
        super().__init__(**response)


# Edit Portfolio
class EditPortfolioResponse(BaseResponse):
    def __init__(self, response: dict):
        if "portfolio" in response:
            self.portfolio: Optional[Portfolio] = Portfolio(**response.pop("portfolio"))
        super().__init__(**response)



import unittest

from requests_mock import Mocker

from coinbase.rest import RESTClient

from ..constants import TEST_API_KEY, TEST_API_SECRET


class PublicTest(unittest.TestCase):
    def test_authenticated_request(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"iso": "2022-01-01T00:00:00Z", "epoch": 1640995200}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/time",
                json=expected_response,
            )
            time = client.get_unix_time()

            captured_request = m.request_history[0]
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.query, "")
            self.assertEqual(time.__dict__, expected_response)
            self.assertIn("Authorization", captured_headers)

    def test_unauthenticated_request(self):
        client = RESTClient()

        expected_response = {"iso": "2022-01-01T00:00:00Z", "epoch": 1640995200}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/time",
                json=expected_response,
            )
            time = client.get_unix_time()

            captured_request = m.request_history[0]
            captured_headers = captured_request.headers

            self.assertEqual(captured_request.query, "")
            self.assertEqual(time.__dict__, expected_response)
            self.assertNotIn("Authorization", captured_headers)

    def test_get_time(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"iso": "2022-01-01T00:00:00Z", "epoch": 1640995200}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/time",
                json=expected_response,
            )
            time = client.get_unix_time()

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(time.__dict__, expected_response)

    def test_get_public_product_book(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/market/product_book",
                json=expected_response,
            )
            book = client.get_public_product_book("product_1", 10)

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "product_id=product_1&limit=10")
            self.assertEqual(book.__dict__, expected_response)

    def test_get_public_products(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/market/products",
                json=expected_response,
            )
            products = client.get_public_products(limit=2, product_type="SPOT")

            captured_request = m.request_history[0]

            self.assertEqual(
                captured_request.query,
                "limit=2&product_type=spot&get_all_products=false",
            )
            self.assertEqual(products.__dict__, expected_response)

    def test_get_public_product(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"product_id": "product_1"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/market/products/product_1",
                json=expected_response,
            )
            product = client.get_public_product("product_1")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(product.__dict__, expected_response)

    def test_get_public_candles(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/market/products/product_id_1/candles",
                json=expected_response,
            )
            candles = client.get_public_candles(
                "product_id_1", "1640995200", "1641081600", "FIVE_MINUTE", 2
            )

            captured_request = m.request_history[0]

            self.assertEqual(
                captured_request.query,
                "start=1640995200&end=1641081600&granularity=five_minute&limit=2",
            )
            self.assertEqual(candles.__dict__, expected_response)

    def test_get_public_market_trades(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/market/products/product_id/ticker",
                json=expected_response,
            )
            trades = client.get_public_market_trades(
                "product_id", 10, "1640995200", "1641081600"
            )

            captured_request = m.request_history[0]

            self.assertEqual(
                captured_request.query, "limit=10&start=1640995200&end=1641081600"
            )
            self.assertEqual(trades.__dict__, expected_response)



from typing import Any, Dict, Optional

from coinbase.constants import API_PREFIX
from coinbase.rest.types.fees_types import GetTransactionSummaryResponse


def get_transaction_summary(
    self,
    product_type: Optional[str] = None,
    contract_expiry_type: Optional[str] = None,
    product_venue: Optional[str] = None,
    **kwargs,
) -> GetTransactionSummaryResponse:
    """
    **Get Transactions Summary**
    _____________________________

    [GET] https://api.coinbase.com/api/v3/brokerage/transaction_summary

    __________

    **Description:**

    Get a summary of transactions with fee tiers, total volume, and fees.

    __________

    **Read more on the official documentation:** `Create Convert Quote <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_createconvertquote>`_
    """
    endpoint = f"{API_PREFIX}/transaction_summary"

    params = {
        "product_type": product_type,
        "contract_expiry_type": contract_expiry_type,
        "product_venue": product_venue,
    }

    return GetTransactionSummaryResponse(self.get(endpoint, params=params, **kwargs))



from typing import List, Optional

from coinbase.websocket.types.base_response import BaseResponse


class WSHeartBeats(BaseResponse):
    def __init__(self, **kwargs):
        self.current_time: Optional[str] = kwargs.pop("current_time", None)
        self.heartbeat_counter: Optional[str] = kwargs.pop("heartbeat_counter", None)
        super().__init__(**kwargs)


class WSCandle(BaseResponse):
    def __init__(self, **kwargs):
        self.start: str = kwargs.pop("start", None)
        self.high: str = kwargs.pop("high", None)
        self.low: str = kwargs.pop("low", None)
        self.open: str = kwargs.pop("open", None)
        self.close: str = kwargs.pop("close", None)
        self.volume: str = kwargs.pop("volume", None)
        self.product_id: str = kwargs.pop("product_id", None)
        super().__init__(**kwargs)


class WSHistoricalMarketTrade(BaseResponse):
    def __init__(self, **kwargs):
        self.product_id: str = kwargs.pop("product_id", None)
        self.trade_id: str = kwargs.pop("trade_id", None)
        self.price: str = kwargs.pop("price", None)
        self.size: str = kwargs.pop("size", None)
        self.time: str = kwargs.pop("time", None)
        self.side: str = kwargs.pop("side", None)
        super().__init__(**kwargs)


class WSProduct(BaseResponse):
    def __init__(self, **kwargs):
        self.product_type: str = kwargs.pop("product_type", None)
        self.id: str = kwargs.pop("id", None)
        self.base_currency: str = kwargs.pop("base_currency", None)
        self.quote_currency: str = kwargs.pop("quote_currency", None)
        self.base_increment: str = kwargs.pop("base_increment", None)
        self.quote_increment: str = kwargs.pop("quote_increment", None)
        self.display_name: str = kwargs.pop("display_name", None)
        self.status: str = kwargs.pop("status", None)
        self.status_message: str = kwargs.pop("status_message", None)
        self.min_market_funds: str = kwargs.pop("min_market_funds", None)
        super().__init__(**kwargs)


class WSTicker(BaseResponse):
    def __init__(self, **kwargs):
        self.type: str = kwargs.pop("type", None)
        self.product_id: str = kwargs.pop("product_id", None)
        self.price: str = kwargs.pop("price", None)
        self.volume_24_h: str = kwargs.pop("volume_24_h", None)
        self.low_24_h: str = kwargs.pop("low_24_h", None)
        self.high_24_h: str = kwargs.pop("high_24_h", None)
        self.low_52_w: str = kwargs.pop("low_52_w", None)
        self.high_52_w: str = kwargs.pop("high_52_w", None)
        self.price_percent_chg_24_h: str = kwargs.pop("price_percent_chg_24_h", None)
        self.best_bid: str = kwargs.pop("best_bid", None)
        self.best_ask: str = kwargs.pop("best_ask", None)
        self.best_bid_quantity: str = kwargs.pop("best_bid_quantity", None)
        self.best_ask_quantity: str = kwargs.pop("best_ask_quantity", None)
        super().__init__(**kwargs)


class L2Update(BaseResponse):
    def __init__(self, **kwargs):
        self.side: str = kwargs.pop("side", None)
        self.event_time: str = kwargs.pop("event_time", None)
        self.price_level: str = kwargs.pop("price_level", None)
        self.new_quantity: str = kwargs.pop("new_quantity", None)
        super().__init__(**kwargs)


class UserOrders(BaseResponse):
    def __init__(self, **kwargs):
        self.avg_price: Optional[str] = kwargs.pop("avg_price", None)
        self.cancel_reason: Optional[str] = kwargs.pop("cancel_reason", None)
        self.client_order_id: Optional[str] = kwargs.pop("client_order_id", None)
        self.completion_percentage: Optional[str] = kwargs.pop(
            "completion_percentage", None
        )
        self.contract_expiry_type: Optional[str] = kwargs.pop(
            "contract_expiry_type", None
        )
        self.cumulative_quantity: Optional[str] = kwargs.pop(
            "cumulative_quantity", None
        )
        self.filled_value: Optional[str] = kwargs.pop("filled_value", None)
        self.leaves_quantity: Optional[str] = kwargs.pop("leaves_quantity", None)
        self.limit_price: Optional[str] = kwargs.pop("limit_price", None)
        self.number_of_fills: Optional[str] = kwargs.pop("number_of_fills", None)
        self.order_id: Optional[str] = kwargs.pop("order_id", None)
        self.order_side: Optional[str] = kwargs.pop("order_side", None)
        self.order_type: Optional[str] = kwargs.pop("order_type", None)
        self.outstanding_hold_amount: Optional[str] = kwargs.pop(
            "outstanding_hold_amount", None
        )
        self.post_only: Optional[str] = kwargs.pop("post_only", None)
        self.product_id: Optional[str] = kwargs.pop("product_id", None)
        self.product_type: Optional[str] = kwargs.pop("product_type", None)
        self.reject_reason: Optional[str] = kwargs.pop("reject_reason", None)
        self.retail_portfolio_id: Optional[str] = kwargs.pop(
            "retail_portfolio_id", None
        )
        self.risk_managed_by: Optional[str] = kwargs.pop("risk_managed_by", None)
        self.status: Optional[str] = kwargs.pop("status", None)
        self.stop_price: Optional[str] = kwargs.pop("stop_price", None)
        self.time_in_force: Optional[str] = kwargs.pop("time_in_force", None)
        self.total_fees: Optional[str] = kwargs.pop("total_fees", None)
        self.total_value_after_fees: Optional[str] = kwargs.pop(
            "total_value_after_fees", None
        )
        self.trigger_status: Optional[str] = kwargs.pop("trigger_status", None)
        self.creation_time: Optional[str] = kwargs.pop("creation_time", None)
        self.end_time: Optional[str] = kwargs.pop("end_time", None)
        self.start_time: Optional[str] = kwargs.pop("start_time", None)
        super().__init__(**kwargs)


class UserPositions(BaseResponse):
    def __init__(self, **kwargs):
        self.perpetual_futures_positions: Optional[List[UserFuturesPositions]] = (
            [
                UserFuturesPositions(**position)
                for position in kwargs.pop("perpetual_futures_positions", [])
            ]
            if kwargs.get("perpetual_futures_positions") is not None
            else []
        )
        self.expiring_futures_positions: Optional[List[UserExpFuturesPositions]] = (
            [
                UserExpFuturesPositions(**position)
                for position in kwargs.pop("expiring_futures_positions", [])
            ]
            if kwargs.get("expiring_futures_positions") is not None
            else []
        )
        super().__init__(**kwargs)


class UserFuturesPositions(BaseResponse):
    def __init__(self, **kwargs):
        self.product_id: Optional[str] = kwargs.pop("product_id", None)
        self.portfolio_uuid: Optional[str] = kwargs.pop("portfolio_uuid", None)
        self.vwap: Optional[str] = kwargs.pop("vwap", None)
        self.entry_vwap: Optional[str] = kwargs.pop("entry_vwap", None)
        self.position_side: Optional[str] = kwargs.pop("position_side", None)
        self.margin_type: Optional[str] = kwargs.pop("margin_type", None)
        self.net_size: Optional[str] = kwargs.pop("net_size", None)
        self.buy_order_size: Optional[str] = kwargs.pop("buy_order_size", None)
        self.sell_order_size: Optional[str] = kwargs.pop("sell_order_size", None)
        self.leverage: Optional[str] = kwargs.pop("leverage", None)
        self.mark_price: Optional[str] = kwargs.pop("mark_price", None)
        self.liquidation_price: Optional[str] = kwargs.pop("liquidation_price", None)
        self.im_notional: Optional[str] = kwargs.pop("im_notional", None)
        self.mm_notional: Optional[str] = kwargs.pop("mm_notional", None)
        self.position_notional: Optional[str] = kwargs.pop("position_notional", None)
        self.unrealized_pnl: Optional[str] = kwargs.pop("unrealized_pnl", None)
        self.aggregated_pnl: Optional[str] = kwargs.pop("aggregated_pnl", None)
        super().__init__(**kwargs)


class UserExpFuturesPositions(BaseResponse):
    def __init__(self, **kwargs):
        self.product_id: Optional[str] = kwargs.pop("product_id", None)
        self.side: Optional[str] = kwargs.pop("side", None)
        self.number_of_contracts: Optional[str] = kwargs.pop(
            "number_of_contracts", None
        )
        self.realized_pnl: Optional[str] = kwargs.pop("realized_pnl", None)
        self.unrealized_pnl: Optional[str] = kwargs.pop("unrealized_pnl", None)
        self.entry_price: Optional[str] = kwargs.pop("entry_price", None)
        super().__init__(**kwargs)


class WSFCMBalanceSummary(BaseResponse):
    def __init__(self, **kwargs):
        self.futures_buying_power: str = kwargs.pop("futures_buying_power", None)
        self.total_usd_balance: str = kwargs.pop("total_usd_balance", None)
        self.cbi_usd_balance: str = kwargs.pop("cbi_usd_balance", None)
        self.cfm_usd_balance: str = kwargs.pop("cfm_usd_balance", None)
        self.total_open_orders_hold_amount: str = kwargs.pop(
            "total_open_orders_hold_amount", None
        )
        self.unrealized_pnl: str = kwargs.pop("unrealized_pnl", None)
        self.daily_realized_pnl: str = kwargs.pop("daily_realized_pnl", None)
        self.initial_margin: str = kwargs.pop("initial_margin", None)
        self.available_margin: str = kwargs.pop("available_margin", None)
        self.liquidation_threshold: str = kwargs.pop("liquidation_threshold", None)
        self.liquidation_buffer_amount: str = kwargs.pop(
            "liquidation_buffer_amount", None
        )
        self.liquidation_buffer_percentage: str = kwargs.pop(
            "liquidation_buffer_percentage", None
        )
        self.intraday_margin_window_measure: Optional[FCMMarginWindowMeasure] = (
            FCMMarginWindowMeasure(**kwargs.pop("intraday_margin_window_measure"))
            if kwargs.get("intraday_margin_window_measure")
            else None
        )
        self.overnight_margin_window_measure: Optional[FCMMarginWindowMeasure] = (
            FCMMarginWindowMeasure(**kwargs.pop("overnight_margin_window_measure"))
            if kwargs.get("overnight_margin_window_measure")
            else None
        )
        super().__init__(**kwargs)


class FCMMarginWindowMeasure(BaseResponse):
    def __init__(self, **kwargs):
        self.margin_window_type: Optional[str] = kwargs.pop("margin_window_type", None)
        self.margin_level: Optional[str] = kwargs.pop("margin_level", None)
        self.initial_margin: Optional[str] = kwargs.pop("initial_margin", None)
        self.maintenance_margin: Optional[str] = kwargs.pop("maintenance_margin", None)
        self.liquidation_buffer_percentage: Optional[str] = kwargs.pop(
            "liquidation_buffer_percentage", None
        )
        self.total_hold: Optional[str] = kwargs.pop("total_hold", None)
        self.futures_buying_power: Optional[str] = kwargs.pop(
            "futures_buying_power", None
        )
        super().__init__(**kwargs)



def object_to_dict(obj):
    if isinstance(obj, list):
        return [object_to_dict(item) for item in obj]
    elif hasattr(obj, "__dict__"):
        return {key: object_to_dict(value) for key, value in obj.__dict__.items()}
    else:
        return obj



import unittest

from requests_mock import Mocker

from coinbase.rest import RESTClient

from ..constants import TEST_API_KEY, TEST_API_SECRET


class PortfoliosTest(unittest.TestCase):
    def test_get_portfolios(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/portfolios",
                json=expected_response,
            )
            portfolios = client.get_portfolios("DEFAULT")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "portfolio_type=default")
            self.assertEqual(portfolios.__dict__, expected_response)

    def test_create_portfolio(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/portfolios",
                json=expected_response,
            )
            portfolio = client.create_portfolio("Test Portfolio")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(captured_json, {"name": "Test Portfolio"})
            self.assertEqual(portfolio.__dict__, expected_response)

    def test_get_portfolio_breakdown(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/1234",
                json=expected_response,
            )
            breakdown = client.get_portfolio_breakdown("1234", "USD")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "currency=usd")
            self.assertEqual(breakdown.__dict__, expected_response)

    def test_move_portfolio_funds(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/move_funds",
                json=expected_response,
            )
            move = client.move_portfolio_funds("100", "USD", "1234", "5678")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "funds": {"value": "100", "currency": "USD"},
                    "source_portfolio_uuid": "1234",
                    "target_portfolio_uuid": "5678",
                },
            )
            self.assertEqual(move.__dict__, expected_response)

    def test_edit_portfolio(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "PUT",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/1234",
                json=expected_response,
            )
            edit = client.edit_portfolio("1234", "Test Portfolio")

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(captured_json, {"name": "Test Portfolio"})
            self.assertEqual(edit.__dict__, expected_response)

    def test_delete_portfolio(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "DELETE",
                "https://api.coinbase.com/api/v3/brokerage/portfolios/1234",
                json=expected_response,
            )
            delete = client.delete_portfolio("1234")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(delete.__dict__, expected_response)



from typing import List, Optional

from coinbase.websocket.types.base_response import BaseResponse
from coinbase.websocket.types.misc_types import (
    L2Update,
    UserOrders,
    UserPositions,
    WSCandle,
    WSFCMBalanceSummary,
    WSHistoricalMarketTrade,
    WSProduct,
    WSTicker,
)


class WebsocketResponse(BaseResponse):
    def __init__(self, data):
        self.channel = data.pop("channel")
        self.client_id = data.pop("client_id")
        self.timestamp = data.pop("timestamp")
        self.sequence_num = data.pop("sequence_num")
        self.events = [
            Event(event_data, self.channel) for event_data in data.pop("events")
        ]

        super().__init__(**data)


class Event(BaseResponse):
    def __init__(self, data, channel):
        if channel == "heartbeats":
            self.current_time = data.pop("current_time", None)
            self.heartbeat_counter = data.pop("heartbeat_counter", None)
        elif channel == "candles":
            self.type = data.pop("type", None)
            self.candles: List[WSCandle] = (
                [WSCandle(**ws_candle) for ws_candle in data.pop("candles", [])]
                if "candles" in data
                else None
            )
        elif channel == "market_trades":
            self.type = data.pop("type", None)
            self.trades: List[WSHistoricalMarketTrade] = (
                [
                    WSHistoricalMarketTrade(**ws_trades)
                    for ws_trades in data.pop("trades", [])
                ]
                if "trades" in data
                else None
            )
        elif channel == "status":
            self.type = data.pop("type", None)
            self.products: List[WSProduct] = (
                [WSProduct(**ws_product) for ws_product in data.pop("products", [])]
                if "products" in data
                else None
            )
        elif channel == "ticker" or channel == "ticker_batch":
            self.type = data.pop("type", None)
            self.tickers: List[WSTicker] = (
                [WSTicker(**ws_ticker) for ws_ticker in data.pop("tickers", [])]
                if "tickers" in data
                else None
            )
        elif channel == "l2_data":
            self.type = data.pop("type", None)
            self.product_id = data.pop("product_id", None)
            self.updates: List[L2Update] = (
                [L2Update(**l2_update) for l2_update in data.pop("updates", [])]
                if "updates" in data
                else None
            )
        elif channel == "user":
            self.type = data.pop("type", None)
            self.orders: Optional[List[UserOrders]] = (
                [UserOrders(**user_order) for user_order in data.pop("orders", [])]
                if data.get("orders") is not None
                else None
            )
            self.positions: Optional[UserPositions] = (
                UserPositions(**data.pop("positions")) if "positions" in data else None
            )
        elif channel == "futures_balance_summary":
            self.type = data.pop("type", None)
            self.fcm_balance_summary: WSFCMBalanceSummary = (
                WSFCMBalanceSummary(**data.pop("fcm_balance_summary"))
                if data.get("fcm_balance_summary")
                else None
            )

        super().__init__(**data)



import os
from typing import IO, Callable, Optional, Union

from coinbase.constants import API_ENV_KEY, API_SECRET_ENV_KEY, WS_USER_BASE_URL

from .types.websocket_response import WebsocketResponse
from .websocket_base import WSBase, WSClientConnectionClosedException, WSClientException


class WSClient(WSBase):
    """
    **WSClient**
    _____________________________

    Initialize using WSClient

    __________

    **Parameters**:

    - **api_key | Optional (str)** - The API key
    - **api_secret | Optional (str)** - The API key secret
    - **key_file | Optional (IO | str)** - Path to API key file or file-like object
    - **base_url | (str)** - The websocket base url. Default set to "wss://advanced-trade-ws.coinbase.com"
    - **timeout | Optional (int)** - Set timeout in seconds for REST requests
    - **max_size | Optional (int)** - Max size in bytes for messages received. Default set to (10 * 1024 * 1024)
    - **on_message | Optional (Callable[[str], None])** - Function called when a message is received
    - **on_open | Optional ([Callable[[], None]])** - Function called when a connection is opened
    - **on_close | Optional ([Callable[[], None]])** - Function called when a connection is closed
    - **retry | Optional (bool)** - Enables automatic reconnections. Default set to True
    - **verbose | Optional (bool)** - Enables debug logging. Default set to False


    """

    from .channels import (
        candles,
        candles_async,
        candles_unsubscribe,
        candles_unsubscribe_async,
        futures_balance_summary,
        futures_balance_summary_async,
        futures_balance_summary_unsubscribe,
        futures_balance_summary_unsubscribe_async,
        heartbeats,
        heartbeats_async,
        heartbeats_unsubscribe,
        heartbeats_unsubscribe_async,
        level2,
        level2_async,
        level2_unsubscribe,
        level2_unsubscribe_async,
        market_trades,
        market_trades_async,
        market_trades_unsubscribe,
        market_trades_unsubscribe_async,
        status,
        status_async,
        status_unsubscribe,
        status_unsubscribe_async,
        ticker,
        ticker_async,
        ticker_batch,
        ticker_batch_async,
        ticker_batch_unsubscribe,
        ticker_batch_unsubscribe_async,
        ticker_unsubscribe,
        ticker_unsubscribe_async,
        user,
        user_async,
        user_unsubscribe,
        user_unsubscribe_async,
    )


class WSUserClient(WSBase):
    """
    **WSUserClient**
    _____________________________

    Initialize using WSUserClient

    __________

    **Parameters**:

    - **api_key | Optional (str)** - The API key
    - **api_secret | Optional (str)** - The API key secret
    - **key_file | Optional (IO | str)** - Path to API key file or file-like object
    - **base_url | (str)** - The websocket base url. Default set to "wss://advanced-trade-ws.coinbase.com"
    - **timeout | Optional (int)** - Set timeout in seconds for REST requests
    - **max_size | Optional (int)** - Max size in bytes for messages received. Default set to (10 * 1024 * 1024)
    - **on_message | Optional (Callable[[str], None])** - Function called when a message is received
    - **on_open | Optional ([Callable[[], None]])** - Function called when a connection is opened
    - **on_close | Optional ([Callable[[], None]])** - Function called when a connection is closed
    - **retry | Optional (bool)** - Enables automatic reconnections. Default set to True
    - **verbose | Optional (bool)** - Enables debug logging. Default set to False


    """

    from .channels import (
        futures_balance_summary,
        futures_balance_summary_async,
        futures_balance_summary_unsubscribe,
        futures_balance_summary_unsubscribe_async,
        heartbeats,
        heartbeats_async,
        heartbeats_unsubscribe,
        heartbeats_unsubscribe_async,
        user,
        user_async,
        user_unsubscribe,
        user_unsubscribe_async,
    )

    def __init__(
        self,
        api_key: Optional[str] = os.getenv(API_ENV_KEY),
        api_secret: Optional[str] = os.getenv(API_SECRET_ENV_KEY),
        key_file: Optional[Union[IO, str]] = None,
        base_url=WS_USER_BASE_URL,
        timeout: Optional[int] = None,
        max_size: Optional[int] = 10 * 1024 * 1024,
        on_message: Optional[Callable[[str], None]] = None,
        on_open: Optional[Callable[[], None]] = None,
        on_close: Optional[Callable[[], None]] = None,
        retry: Optional[bool] = True,
        verbose: Optional[bool] = False,
    ):
        super().__init__(
            api_key=api_key,
            api_secret=api_secret,
            key_file=key_file,
            base_url=base_url,
            timeout=timeout,
            max_size=max_size,
            on_message=on_message,
            on_open=on_open,
            on_close=on_close,
            retry=retry,
            verbose=verbose,
        )



from typing import List

from coinbase.constants import (
    CANDLES,
    FUTURES_BALANCE_SUMMARY,
    HEARTBEATS,
    LEVEL2,
    MARKET_TRADES,
    STATUS,
    TICKER,
    TICKER_BATCH,
    USER,
)


def heartbeats(self) -> None:
    """
    **Heartbeats Subscribe**
    ________________________

    __________

    **Description:**

    Subscribe to heartbeats channel.

    __________

    **Read more on the official documentation:** `Heartbeats Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#heartbeats-channel>`_
    """
    self.subscribe([], [HEARTBEATS])


async def heartbeats_async(self) -> None:
    """
    **Heartbeats Subscribe Async**
    ______________________________

    __________

    **Description:**

    Async subscribe to heartbeats channel.

    __________

    **Read more on the official documentation:** `Heartbeats Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#heartbeats-channel>`_
    """
    await self.subscribe_async([], [HEARTBEATS])


def heartbeats_unsubscribe(self) -> None:
    """
    **Heartbeats Unsubscribe**
    __________________________

    __________

    **Description:**

    Unsubscribe to heartbeats channel.

    __________

    **Read more on the official documentation:** `Heartbeats Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#heartbeats-channel>`_
    """
    self.unsubscribe([], [HEARTBEATS])


async def heartbeats_unsubscribe_async(
    self,
) -> None:
    """
    **Heartbeats Unsubscribe Async**
    ________________________________

    __________

    **Description:**

    Async unsubscribe to heartbeats channel.

    __________

    **Read more on the official documentation:** `Heartbeats Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#heartbeats-channel>`_
    """
    await self.unsubscribe_async([], [HEARTBEATS])


def candles(self, product_ids: List[str]) -> None:
    """
    **Candles Subscribe**
    _____________________

    __________

    **Description:**

    Subscribe to candles channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Candles Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#candles-channel>`_
    """
    self.subscribe(product_ids, [CANDLES])


async def candles_async(self, product_ids: List[str]) -> None:
    """
    **Candles Subscribe Async**
    ___________________________

    __________

    **Description:**

    Async subscribe to candles channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Candles Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#candles-channel>`_
    """
    await self.subscribe_async(product_ids, [CANDLES])


def candles_unsubscribe(self, product_ids: List[str]) -> None:
    """
    **Candles Unsubscribe**
    _______________________

    __________

    **Description:**

    Unsubscribe to candles channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Candles Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#candles-channel>`_
    """
    self.unsubscribe(product_ids, [CANDLES])


async def candles_unsubscribe_async(self, product_ids: List[str]) -> None:
    """
    **Candles Unsubscribe Async**
    _____________________________

    __________

    **Description:**

    Async unsubscribe to candles channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Candles Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#candles-channel>`_
    """
    await self.unsubscribe_async(product_ids, [CANDLES])


def market_trades(self, product_ids: List[str]) -> None:
    """
    **Market Trades Subscribe**
    ___________________________

    __________

    **Description:**

    Subscribe to market_trades channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Market Trades Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#market-trades-channel>`_
    """
    self.subscribe(product_ids, [MARKET_TRADES])


async def market_trades_async(self, product_ids: List[str]) -> None:
    """
    **Market Trades Subscribe Async**
    _________________________________

    __________

    **Description:**

    Async subscribe to market_trades channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Market Trades Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#market-trades-channel>`_
    """
    await self.subscribe_async(product_ids, [MARKET_TRADES])


def market_trades_unsubscribe(self, product_ids: List[str]) -> None:
    """
    **Market Trades Unsubscribe**
    _____________________________

    __________

    **Description:**

    Unsubscribe to market_trades channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Market Trades Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#market-trades-channel>`_
    """
    self.unsubscribe(product_ids, [MARKET_TRADES])


async def market_trades_unsubscribe_async(self, product_ids: List[str]) -> None:
    """
    **Market Trades Unsubscribe Async**
    ___________________________________

    __________

    **Description:**

    Async unsubscribe to market_trades channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Market Trades Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#market-trades-channel>`_
    """
    await self.unsubscribe_async(product_ids, [MARKET_TRADES])


def status(self, product_ids: List[str]) -> None:
    """
    **Status Subscribe**
    ____________________

    __________

    **Description:**

    Subscribe to status channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Status Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#status-channel>`_
    """
    self.subscribe(product_ids, [STATUS])


async def status_async(self, product_ids: List[str]) -> None:
    """
    **Status Subscribe Async**
    __________________________

    __________

    **Description:**

    Async subscribe to status channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Status Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#status-channel>`_
    """
    await self.subscribe_async(product_ids, [STATUS])


def status_unsubscribe(self, product_ids: List[str]) -> None:
    """
    **Status Unsubscribe**
    ______________________

    __________

    **Description:**

    Unsubscribe to status channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Status Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#status-channel>`_
    """
    self.unsubscribe(product_ids, [STATUS])


async def status_unsubscribe_async(self, product_ids: List[str]) -> None:
    """
    **Status Unsubscribe Async**
    ____________________________

    __________

    **Description:**

    Async unsubscribe to status channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Status Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#status-channel>`_
    """
    await self.unsubscribe_async(product_ids, [STATUS])


def ticker(self, product_ids: List[str]) -> None:
    """
    **Ticker Subscribe**
    ____________________

    __________

    **Description:**

    Subscribe to ticker channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-channel>`_
    """
    self.subscribe(product_ids, [TICKER])


async def ticker_async(self, product_ids: List[str]) -> None:
    """
    **Ticker Subscribe Async**
    __________________________

    __________

    **Description:**

    Async subscribe to ticker channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-channel>`_
    """
    await self.subscribe_async(product_ids, [TICKER])


def ticker_unsubscribe(self, product_ids: List[str]) -> None:
    """
    **Ticker Unsubscribe**
    ______________________

    __________

    **Description:**

    Unsubscribe to ticker channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-channel>`_
    """
    self.unsubscribe(product_ids, [TICKER])


async def ticker_unsubscribe_async(self, product_ids: List[str]) -> None:
    """
    **Ticker Unsubscribe Async**
    ____________________________

    __________

    **Description:**

    Async unsubscribe to ticker channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-channel>`_
    """
    await self.unsubscribe_async(product_ids, [TICKER])


def ticker_batch(self, product_ids: List[str]) -> None:
    """
    **Ticker Batch Subscribe**
    __________________________

    __________

    **Description:**

    Subscribe to ticker_batch channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Batch Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-batch-channel>`_
    """
    self.subscribe(product_ids, [TICKER_BATCH])


async def ticker_batch_async(self, product_ids: List[str]) -> None:
    """
    **Ticker Batch Subscribe Async**
    ________________________________

    __________

    **Description:**

    Async subscribe to ticker_batch channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Batch Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-batch-channel>`_
    """
    await self.subscribe_async(product_ids, [TICKER_BATCH])


def ticker_batch_unsubscribe(self, product_ids: List[str]) -> None:
    """
    **Ticker Batch Unsubscribe**
    ____________________________

    __________

    **Description:**

    Unsubscribe to ticker_batch channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Batch Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-batch-channel>`_
    """
    self.unsubscribe(product_ids, [TICKER_BATCH])


async def ticker_batch_unsubscribe_async(self, product_ids: List[str]) -> None:
    """
    **Ticker Batch Unsubscribe Async**
    __________________________________

    __________

    **Description:**

    Async unsubscribe to ticker_batch channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Ticker Batch Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#ticker-batch-channel>`_
    """
    await self.unsubscribe_async(product_ids, [TICKER_BATCH])


def level2(self, product_ids: List[str]) -> None:
    """
    **Level2 Subscribe**
    ____________________

    __________

    **Description:**

    Subscribe to level2 channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Level2 Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#level2-channel>`_
    """
    self.subscribe(product_ids, [LEVEL2])


async def level2_async(self, product_ids: List[str]) -> None:
    """
    **Level2 Subscribe Async**
    __________________________

    __________

    **Description:**

    Async subscribe to level2 channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Level2 Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#level2-channel>`_
    """
    await self.subscribe_async(product_ids, [LEVEL2])


def level2_unsubscribe(self, product_ids: List[str]) -> None:
    """
    **Level2 Unsubscribe**
    ______________________

    __________

    **Description:**

    Unsubscribe to level2 channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Level2 Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#level2-channel>`_
    """
    self.unsubscribe(product_ids, [LEVEL2])


async def level2_unsubscribe_async(self, product_ids: List[str]) -> None:
    """
    **Level2 Unsubscribe Async**
    ____________________________

    __________

    **Description:**

    Async unsubscribe to level2 channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `Level2 Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#level2-channel>`_
    """
    await self.unsubscribe_async(product_ids, [LEVEL2])


def user(self, product_ids: List[str]) -> None:
    """
    **User Subscribe**
    __________________

    __________

    **Description:**

    Subscribe to user channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `User Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#user-channel>`_
    """
    self.subscribe(product_ids, [USER])


async def user_async(self, product_ids: List[str]) -> None:
    """
    **User Subscribe Async**
    ________________________

    __________

    **Description:**

    Async subscribe to user channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `User Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#user-channel>`_
    """
    await self.subscribe_async(product_ids, [USER])


def user_unsubscribe(self, product_ids: List[str]) -> None:
    """
    **User Unsubscribe**
    ____________________

    __________

    **Description:**

    Unsubscribe to user channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `User Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#user-channel>`_
    """
    self.unsubscribe(product_ids, [USER])


async def user_unsubscribe_async(self, product_ids: List[str]) -> None:
    """
    **User Unsubscribe Async**
    __________________________

    __________

    **Description:**

    Async unsubscribe to user channel for a list of products_ids.

    __________

    **Read more on the official documentation:** `User Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#user-channel>`_
    """
    await self.unsubscribe_async(product_ids, [USER])


def futures_balance_summary(self) -> None:
    """
    **Futures Balance Summary Subscribe**
    __________________

    __________

    **Description:**

    Subscribe to futures_balance_summary channel.

    __________

    **Read more on the official documentation:** `Futures Balance Summary Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#futures-balance-summary-channel>`_
    """
    self.subscribe([], [FUTURES_BALANCE_SUMMARY])


async def futures_balance_summary_async(self) -> None:
    """
    **Futures Balance Summary Subscribe Async**
    ________________________

    __________

    **Description:**

    Async subscribe to futures_balance_summary channel.

    __________

    **Read more on the official documentation:** `Futures Balance Summary Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#futures-balance-summary-channel>`_
    """
    await self.subscribe_async([], [FUTURES_BALANCE_SUMMARY])


def futures_balance_summary_unsubscribe(self) -> None:
    """
    **Futures Balance Summary Unsubscribe**
    ____________________

    __________

    **Description:**

    Unsubscribe to futures_balance_summary channel.

    __________

    **Read more on the official documentation:** `Futures Balance Summary Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#futures-balance-summary-channel>`_
    """
    self.unsubscribe([], [FUTURES_BALANCE_SUMMARY])


async def futures_balance_summary_unsubscribe_async(self) -> None:
    """
    **Futures Balance Summary Unsubscribe Async**
    __________________________

    __________

    **Description:**

    Async unsubscribe to futures_balance_summary channel.

    __________

    **Read more on the official documentation:** `Futures Balance Summary Channel
    <https://docs.cdp.coinbase.com/advanced-trade/docs/ws-channels#futures-balance-summary-channel>`_
    """
    await self.unsubscribe_async([], [FUTURES_BALANCE_SUMMARY])



from typing import Any, Dict, List, Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import (
    CancelOrderObject,
    Order,
    OrderConfiguration,
)


# Create Order
class CreateOrderResponse(BaseResponse):
    def __init__(self, response: dict):
        if "success" in response:
            self.success: bool = response.pop("success")
        if "failure_reason" in response:
            self.failure_reason: Optional[Dict[str, Any]] = response.pop(
                "failure_reason"
            )
        if "order_id" in response:
            self.order_id: Optional[str] = response.pop("order_id")

        if "success_response" in response:
            self.response: Optional[Dict[str, Any]] = response.pop("success_response")
        elif "error_response" in response:
            self.response: Optional[Dict[str, Any]] = response.pop("error_response")

        if "order_configuration" in response:
            self.order_configuration: Optional[OrderConfiguration] = OrderConfiguration(
                **response.pop("order_configuration")
            )
        super().__init__(**response)


# Cancel Orders
class CancelOrdersResponse(BaseResponse):
    def __init__(self, response: dict):
        if "results" in response:
            self.results: Optional[List[CancelOrderObject]] = [
                CancelOrderObject(**result) for result in response.pop("results")
            ]
        super().__init__(**response)


# Edit Order
class EditOrderResponse(BaseResponse):
    def __init__(self, response: dict):
        if "success" in response:
            self.success: bool = response.pop("success")

        if "success_response" in response:
            self.response: Optional[Dict[str, Any]] = response.pop("success_response")
        elif "error_response" in response:
            self.response: Optional[Dict[str, Any]] = response.pop("error_response")

        if "errors" in response:
            self.errors: Optional[List[Dict[str, Any]]] = response.pop("errors")
        super().__init__(**response)


# Edit Order Preview
class EditOrderPreviewResponse(BaseResponse):
    def __init__(self, response: dict):
        if "errors" in response:
            self.errors: List[Dict[str, Any]] = response.pop("errors")
        if "slippage" in response:
            self.slippage: Optional[str] = response.pop("slippage")
        if "order_total" in response:
            self.order_total: Optional[str] = response.pop("order_total")
        if "commission_total" in response:
            self.commission_total: Optional[str] = response.pop("commission_total")
        if "quote_size" in response:
            self.quote_size: Optional[str] = response.pop("quote_size")
        if "base_size" in response:
            self.base_size: Optional[str] = response.pop("base_size")
        if "best_bid" in response:
            self.best_bid: Optional[str] = response.pop("best_bid")
        if "average_filled_price" in response:
            self.average_filled_price: Optional[str] = response.pop(
                "average_filled_price"
            )
        super().__init__(**response)


# List Orders
class ListOrdersResponse(BaseResponse):
    def __init__(self, response: dict):
        if "orders" in response:
            self.orders: List[Order] = [
                Order(**order) for order in response.pop("orders")
            ]
        if "sequence" in response:
            self.sequence: Optional[int] = response.pop("sequence")
        if "has_next" in response:
            self.has_next: bool = response.pop("has_next")
        if "cursor" in response:
            self.cursor: Optional[str] = response.pop("cursor")
        super().__init__(**response)


# List Fills
class ListFillsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "fills" in response:
            self.fills: Optional[List[Dict[str, Any]]] = response.pop("fills")
        if "cursor" in response:
            self.cursor: Optional[str] = response.pop("cursor")
        super().__init__(**response)


# Get Order
class GetOrderResponse(BaseResponse):
    def __init__(self, response: dict):
        if "order" in response:
            self.order: Optional[Order] = Order(**response.pop("order"))
        super().__init__(**response)


# Preview Order
class PreviewOrderResponse(BaseResponse):
    def __init__(self, response: dict):
        if "order_total" in response:
            self.order_total: str = response.pop("order_total")
        if "commission_total" in response:
            self.commission_total: str = response.pop("commission_total")
        if "errs" in response:
            self.errs: List[Dict[str, Any]] = response.pop("errs")
        if "warning" in response:
            self.warning: List[Dict[str, Any]] = response.pop("warning")
        if "quote_size" in response:
            self.quote_size: str = response.pop("quote_size")
        if "base_size" in response:
            self.base_size: str = response.pop("base_size")
        if "best_bid" in response:
            self.best_bid: str = response.pop("best_bid")
        if "best_ask" in response:
            self.best_ask: str = response.pop("best_ask")
        if "is_max" in response:
            self.is_max: bool = response.pop("is_max")
        if "order_margin_total" in response:
            self.order_margin_total: Optional[str] = response.pop("order_margin_total")
        if "leverage" in response:
            self.leverage: Optional[str] = response.pop("leverage")
        if "long_leverage" in response:
            self.long_leverage: Optional[str] = response.pop("long_leverage")
        if "short_leverage" in response:
            self.short_leverage: Optional[str] = response.pop("short_leverage")
        if "slippage" in response:
            self.slippage: Optional[str] = response.pop("slippage")
        if "preview_id" in response:
            self.preview_id: Optional[str] = response.pop("preview_id")
        if "current_liquidation_buffer" in response:
            self.current_liquidation_buffer: Optional[str] = response.pop(
                "current_liquidation_buffer"
            )
        if "projected_liquidation_buffer" in response:
            self.projected_liquidation_buffer: Optional[str] = response.pop(
                "projected_liquidation_buffer"
            )
        if "max_leverage" in response:
            self.max_leverage: Optional[str] = response.pop("max_leverage")
        if "pnl_configuration" in response:
            self.pnl_configuration: Optional[Dict[str, Any]] = response.pop(
                "pnl_configuration"
            )
        super().__init__(**response)


# Close Position
class ClosePositionResponse(BaseResponse):
    def __init__(self, response: dict):
        if "success" in response:
            self.success: bool = response.pop("success")

        if "success_response" in response:
            self.response: Optional[Dict[str, Any]] = response.pop("success_response")
        elif "error_response" in response:
            self.response: Optional[Dict[str, Any]] = response.pop("error_response")

        if "order_configuration" in response:
            self.order_configuration: Optional[OrderConfiguration] = OrderConfiguration(
                **response.pop("order_configuration")
            )
        super().__init__(**response)
